/*
 * FileReader.cpp
 *
 *  Created on: Nov 11, 2013
 *      Author: mswindell
 */

#include "FileReader.h"
#include "Interpolator.h"
#include "TimeInterpolator.h"
#include "cdf.h"
#include "Time.h"
#include <fstream>
#include <iterator>
#include <algorithm>
#include <vector>
#include <math.h>

FileReader::FileReader(int lat, int lon, std::string variable, std::string filename) {
    this->lat = lat;
    this->lon = lon;
    this->variable = variable;
    this->filename = filename;
}

FileReader::FileReader() {

}

void FileReader::close() {
    //TODO: Write closing code
    if (variableDataMap.size() > 0) {
        std::map<std::string, std::vector<float>*>::iterator iter;

        for (iter = variableDataMap.begin(); iter != variableDataMap.end(); iter++) {
            //std::cout << "delete " << iter->first << ", ";
            delete iter->second;
        }
        //std::cout << "\n";
    }
    //the other maps have the same pointers, so they are already deleted.
    variableDataMap.clear();
    loadedVariables.clear();
    variableIds.clear();
    variableNames.clear();
    loadedVariablesById.clear();
    gAttributes.clear();
    gAttributeById.clear();
    CDFcloseCDF(currentFileId);
}

bool FileReader::doesVariableExist(const std::string& variable) {
    if (variableIds.find(variable) != variableIds.end())
        return true;
    return false;
}

CDFid FileReader::getCurrentFileId() {
    return currentFileId;
}

std::string FileReader::getCurrentFilename() {
    return currentFilename;
}

Time FileReader::getCurrentTime() {
    Attribute attribute = getGlobalAttribute("timestep_time");
    Time time = TimeInterpolator::parseTime(attribute.getAttributeString());
    return time;
}

Attribute FileReader::getGlobalAttribute(long id) {
    //first check if the attribute has already been requested. If so, return stored value
    if (gAttributeById.find(id) != gAttributeById.end()) {
        return gAttributeById[id];
    }

    long attrNum = id;
    long numElements;
    long dataType;
    CDFstatus status = CDFgetAttrgEntryDataType(this->currentFileId, attrNum, 0, &dataType);
    status = CDFgetAttrgEntryNumElements(this->currentFileId, attrNum, 0, &numElements);

    Attribute attribute;

    if (dataType == CDF_CHAR) {
        std::string attributeValue = "NULL";
        char attributeBuffer[numElements + 1];
        CDFgetAttrgEntry(currentFileId, attrNum, 0, attributeBuffer);
        attributeBuffer[numElements] = '\0';
        attributeValue = attributeBuffer;
        attributeValue = attributeValue.substr(0, numElements);

        char attributeNameBuffer[512];

        CDFgetAttrName(currentFileId, attrNum, attributeNameBuffer);
        //std::cout << "attrNum: " << attrNum << " i: " << i << " numElements: " << numElements << std::endl;
        //Attribute attribute;
        attribute.setAttributeName(attributeNameBuffer);
        //std::cout << "attributeBuffer: " << attributeBuffer << endl;
        attribute.setAttributeValue(attributeValue);
        gAttributeById[id] = attribute;
        gAttributes[attribute.getAttributeName()] = attribute;
    } else if (dataType == CDF_INT4) {
        int attributeBuffer;

        CDFgetAttrgEntry(currentFileId, attrNum, 0, (void*) &attributeBuffer);
        //std::cout << "attrNum: " << attrNum << " i: " << i << " numElements: " << numElements << std::endl;
        //std::cout << "numElements: " << numElements << std::endl;
        //std::cout << "status: " << status << std::endl;

        char attributeNameBuffer[512];

        CDFgetAttrName(currentFileId, attrNum, attributeNameBuffer);
        //Attribute attribute;
        attribute.setAttributeName(attributeNameBuffer);
        attribute.setAttributeValue(attributeBuffer);
        gAttributeById[id] = attribute;
        gAttributes[attribute.getAttributeName()] = attribute;
    } else if (dataType == CDF_FLOAT) {
        float attributeBuffer;

        CDFgetAttrgEntry(currentFileId, attrNum, 0, &attributeBuffer);
        //std::cout << "numElements: " << numElements << std::endl;
        //std::cout << "status: " << status << std::endl;

        char attributeNameBuffer[1024];
        CDFgetAttrName(currentFileId, attrNum, attributeNameBuffer);
        //Attribute attribute;
        attribute.setAttributeName(attributeNameBuffer);
        attribute.setAttributeValue(attributeBuffer);
        gAttributeById[id] = attribute;
        gAttributes[attribute.getAttributeName()] = attribute;
        //return attribute;
    }
    //cout << "added: " << i << " name: " << attribute.getAttributeName() << endl;
    //std::cout << "Attribute: " << attribute.toString() << std::endl;
    return attribute;
}

Attribute FileReader::getGlobalAttribute(const std::string& attribute) {
    //first check if the attribute has already been requested. If so, return stored value
    if (gAttributes.find(attribute) != gAttributes.end()) {
        return gAttributes[attribute];
    }
    long attrNum = getGlobalAttributeID(attribute);
    Attribute current_attribute;
    if (attrNum < 0) {
        //std::cout << "attrNum: " << attrNum << " returned for " << attribute << std::endl;
    } else {
        current_attribute = getGlobalAttribute(attrNum);
        //gAttributeById[attrNum] = current_attribute;
        //gAttributes[attribute] = current_attribute;
        return current_attribute;
    }
}

long FileReader::getGlobalAttributeID(const std::string& attribute) {
    long attrNum = CDFgetAttrNum(currentFileId, (char*) attribute.c_str());
    return attrNum;
}

std::vector<std::string> FileReader::getLoadedVariables() {
    return this->loadedVariables;
}

std::vector<std::string> FileReader::getListOfRequiredVariables(std::string variable) {
    if (doesVariableExist(variable)) {
        std::vector<std::string> required;
        required.push_back(variable);
        return required;
    } else {
        std::vector<std::string> required;
        return required;
    }
}

float FileReader::getMissingValue() {
    return missingValue;
}

std::vector<float>* FileReader::getVariableData(const std::string variable) {
    std::vector<float>* varData = NULL;
    if (variableDataMap.find(variable) == variableDataMap.end()) {
        std::cout << "Variable" << variable << " is not in the map!" << std::endl;

    } else {
        varData = variableDataMap[variable];
    }
    return varData;
}

long FileReader::getVariableId(const std::string& variable) {
    if (variableIds.find(variable) != variableIds.end()) {
        return variableIds[variable];
    } else
        return -1;
}

void FileReader::initializeVariableMaps() {
    long numVars;
    CDFgetNumzVars(currentFileId, &numVars);
    char variableName[512];
    for (int i = 0; i < (int) numVars; i++) {
        CDFgetzVarName(currentFileId, i, variableName);
        std::string variableNameString = variableName;
        variableIds[variableNameString] = i;
        variableNames[(long) i] = variableNameString;
    }
}

CDFstatus FileReader::loadVariable(const std::string& variable) {
    //Get the variable id
    long variableNum = CDFgetVarNum(currentFileId, (char *) variable.c_str());
    //Check to see if the variable exists
    if (variableNum < 0) {
        return -1;
    } else {
        //Get the variable dimensions
        long recStart = 0L;
        long recCount = 1L;
        long recInterval = 1L;
        long dimIndices[] = {0};
        long dimIntervals[] = {1};
        long dimSize[1];

        //Get the dimension size
        CDFgetzVarDimSizes(currentFileId, variableNum, dimSize);
        //Get the variable data type
        long dataType;
        CDFgetzVarDataType(currentFileId, variableNum, &dataType);
        std::vector<float>* variableData = new std::vector<float>();
        //std::cout << "new load: " << variable << std::endl;
        variableData->clear();
        //Load variable data
        if (dataType == CDF_FLOAT) {
            float* buffer = new float[dimSize[0]];
            CDFhyperGetzVarData(currentFileId, variableNum, recStart, recCount, recInterval, dimIndices, dimSize,
                    dimIntervals, buffer);
            //add data to vector type, and delete original array
            variableData->reserve(dimSize[0]);
            for (int i = 0; i < dimSize[0]; i++) {
                variableData->push_back(buffer[i]);
            }

            delete[] buffer;
            loadedVariables.push_back(variable);
            loadedVariablesById.push_back(variableNum);
            variableDataMap[variable] = variableData;
        } else if (dataType == CDF_DOUBLE) {
            double* buffer = new double[dimSize[0]];
            CDFhyperGetzVarData(currentFileId, variableNum, recStart, recCount, recInterval, dimIndices, dimSize,
                    dimIntervals, buffer);
            variableData->reserve(dimSize[0]);
            for (int i = 0; i < dimSize[0]; i++) {
                variableData->push_back((float) buffer[i]);

            }
            delete[] buffer;
            loadedVariables.push_back(variable);
            loadedVariablesById.push_back(variableNum);
            variableDataMap[variable] = variableData;

        }
    }
    return 0;
}

long FileReader::open(const std::string& filename) {
    CDFid id;
    CDFstatus status;

    if (!(std::ifstream(filename.c_str()))) {
        status = -2;
        std::cerr << "Filename: \"" << filename << "\" does not exist." << std::endl;
    } else {
        status = CDFopenCDF(filename.c_str(), &id);
        if (status == CDF_OK) {
            setCurrentFileId(id);
            currentFilename = filename;
        }
    }
    loadVariable("theta");
    loadVariable("phi");
    initializeVariableMaps();

    return status;
}

void FileReader::setCurrentFileId(CDFid currentFileId) {
    this->currentFileId = currentFileId;
}

bool FileReader::unloadVariable(const std::string& variable) {
    std::vector<std::string> requiredVariables = this->getListOfRequiredVariables(variable);
    bool success = true;
    for (int i = 0; i < requiredVariables.size(); i++) {
        if (variableDataMap.find(variable) != variableDataMap.end()) {
            std::vector<float> * data = variableDataMap[variable];
            long id = getVariableId(variable);
            delete data;
            //std::cout << "Actually unloading a variable: " << variable << std::endl;
            data = NULL;
            variableDataMap.erase(variable);
            //std::cout << "Got here." << std::endl;
            std::vector<std::string>::iterator it = std::find(loadedVariables.begin(), loadedVariables.end(), variable);
            loadedVariables.erase(it);
        } else
            return false;
    }
    return success;
}

FileReader::~FileReader() {
    // TODO Auto-generated destructor stub
}
